@@ -3,18 +3,26 @@
 import io
 import os
 from pathlib import Path
+from typing import Any, cast
 
 import httpx
 
 from entitysdk import core, route
 from entitysdk.common import ProjectContext
 from entitysdk.exception import EntitySDKError
-from entitysdk.models.asset import Asset, LocalAssetMetadata
+from entitysdk.models.asset import Asset, DetailedFileList, LocalAssetMetadata
 from entitysdk.models.core import Identifiable
+from entitysdk.models.entity import Entity
 from entitysdk.result import IteratorResult
-from entitysdk.token_manager import TokenManager
-from entitysdk.types import ID, DeploymentEnvironment
-from entitysdk.util import build_api_url
+from entitysdk.schemas.asset import DownloadedAsset
+from entitysdk.token_manager import TokenFromValue, TokenManager
+from entitysdk.types import ID, DeploymentEnvironment, Token
+from entitysdk.util import (
+    build_api_url,
+    create_intermediate_directories,
+    validate_filename_extension_consistency,
+)
+from entitysdk.utils.asset import filter_assets
 
 
 class Client:
@@ -22,10 +30,11 @@ class Client:
 
     def __init__(
         self,
+        *,
         api_url: str | None = None,
         project_context: ProjectContext | None = None,
         http_client: httpx.Client | None = None,
-        token_manager: TokenManager | None = None,
+        token_manager: TokenManager | Token,
         environment: DeploymentEnvironment | str | None = None,
     ) -> None:
         """Initialize client.
@@ -34,7 +43,7 @@ class Client:
             api_url: The API URL to entitycore service.
             project_context: Project context.
             http_client: Optional HTTP client to use.
-            token_manager: Optional token manager to use.
+            token_manager: Token manager or token to be used for authentication.
             environment: Deployment environent.
         """
         try:
@@ -51,7 +60,9 @@ class Client:
         )
         self.project_context = project_context
         self._http_client = http_client or httpx.Client()
-        self._token_manager = token_manager
+        self._token_manager = (
+            TokenFromValue(token_manager) if isinstance(token_manager, Token) else token_manager
+        )
 
     @staticmethod
     def _handle_api_url(api_url: str | None, environment: DeploymentEnvironment | None) -> str:
@@ -68,21 +79,6 @@ class Client:
             case _:
                 raise EntitySDKError("Either api_url or environment is of the wrong type.")
 
-    def _get_token(self, override_token: str | None = None) -> str:
-        """Get a token either from an override or from the token manager.
-
-        Args:
-            override_token: Optional override token.
-
-        Returns:
-            Token.
-        """
-        if override_token:
-            return override_token
-        if self._token_manager is None:
-            raise EntitySDKError("Either override_token or token_manager must be provided.")
-        return self._token_manager.get_token()
-
     def _optional_user_context(
         self, override_context: ProjectContext | None
     ) -> ProjectContext | None:
@@ -99,18 +95,14 @@ class Client:
         entity_id: ID,
         *,
         entity_type: type[Identifiable],
-        with_assets: bool = True,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Identifiable:
         """Get entity from resource id.
 
         Args:
             entity_id: Resource id of the entity.
             entity_type: Type of the entity.
-            with_assets: Whether to include assets in the response.
             project_context: Optional project context.
-            token: Authorization access token.
 
         Returns:
             entity_type instantiated by deserializing the response.
@@ -120,30 +112,14 @@ class Client:
             entity_type=entity_type,
             entity_id=entity_id,
         )
-        token = self._get_token(override_token=token)
         context = self._optional_user_context(override_context=project_context)
-        entity = core.get_entity(
+        return core.get_entity(
             url=url,
-            token=token,
             entity_type=entity_type,
             project_context=context,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
-        if with_assets and "assets" in entity_type.model_fields:
-            url = route.get_assets_endpoint(
-                api_url=self.api_url,
-                entity_type=entity_type,
-                entity_id=entity_id,
-            )
-            assets = core.get_entity_assets(
-                url=url,
-                token=token,
-                project_context=context,
-                http_client=self._http_client,
-            )
-            entity = entity.evolve(assets=assets)
-
-        return entity
 
     def search_entity(
         self,
@@ -152,7 +128,6 @@ class Client:
         query: dict | None = None,
         limit: int | None = None,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> IteratorResult[Identifiable]:
         """Search for entities.
 
@@ -161,19 +136,17 @@ class Client:
             query: Query parameters.
             limit: Optional limit of the number of entities to yield. Default is None.
             project_context: Optional project context.
-            token: Authorization access token.
         """
         url = route.get_entities_endpoint(api_url=self.api_url, entity_type=entity_type)
-        token = self._get_token(override_token=token)
         context = self._optional_user_context(override_context=project_context)
         return core.search_entities(
             url=url,
             query=query,
             limit=limit,
-            token=token,
             project_context=context,
             entity_type=entity_type,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def register_entity(
@@ -181,27 +154,24 @@ class Client:
         entity: Identifiable,
         *,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Identifiable:
         """Register entity.
 
         Args:
             entity: Identifiable to register.
             project_context: Optional project context.
-            token: Authorization access token.
 
         Returns:
             Registered entity with id.
         """
         url = route.get_entities_endpoint(api_url=self.api_url, entity_type=type(entity))
         context = self._required_user_context(override_context=project_context)
-        token = self._get_token(override_token=token)
         return core.register_entity(
             url=url,
-            token=token,
             entity=entity,
             project_context=context,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def update_entity(
@@ -211,7 +181,6 @@ class Client:
         attrs_or_entity: dict | Identifiable,
         *,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Identifiable:
         """Update an entity.
 
@@ -220,22 +189,20 @@ class Client:
             entity_type: Type of the entity.
             attrs_or_entity: Attributes or entity to update.
             project_context: Optional project context.
-            token: Authorization access token.
         """
         url = route.get_entities_endpoint(
             api_url=self.api_url,
             entity_type=entity_type,
             entity_id=entity_id,
         )
-        token = self._get_token(override_token=token)
         context = self._required_user_context(override_context=project_context)
         return core.update_entity(
             url=url,
-            token=token,
             project_context=context,
             entity_type=entity_type,
             attrs_or_entity=attrs_or_entity,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def upload_file(
@@ -248,7 +215,6 @@ class Client:
         file_name: str | None = None,
         file_metadata: dict | None = None,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Asset:
         """Upload asset to an existing entity's endpoint from a file path."""
         path = Path(file_path)
@@ -258,7 +224,6 @@ class Client:
             entity_id=entity_id,
             asset_id=None,
         )
-        token = self._get_token(override_token=token)
         context = self._required_user_context(override_context=project_context)
         asset_metadata = LocalAssetMetadata(
             file_name=file_name or path.name,
@@ -267,11 +232,11 @@ class Client:
         )
         return core.upload_asset_file(
             url=url,
-            token=token,
             asset_path=path,
             project_context=context,
             asset_metadata=asset_metadata,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def upload_content(
@@ -284,7 +249,6 @@ class Client:
         file_content_type: str,
         file_metadata: dict | None = None,
         project_context: ProjectContext | None = None,
-        token: str,
     ) -> Asset:
         """Upload asset to an existing entity's endpoint from a file-like object."""
         url = route.get_assets_endpoint(
@@ -298,25 +262,144 @@ class Client:
             content_type=file_content_type,
             metadata=file_metadata or {},
         )
-        token = self._get_token(override_token=token)
         context = self._required_user_context(override_context=project_context)
         return core.upload_asset_content(
             url=url,
-            token=token,
             project_context=context,
             asset_content=file_content,
             asset_metadata=asset_metadata,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
+    def upload_directory(
+        self,
+        *,
+        entity_id: ID,
+        entity_type: type[Identifiable],
+        name: str,
+        paths: dict[os.PathLike, os.PathLike],
+        metadata: dict | None = None,
+        label: str | None = None,
+        project_context: ProjectContext | None = None,
+    ) -> Asset:
+        """Attach directory to an entity from with a group of paths."""
+        url = (
+            route.get_assets_endpoint(
+                api_url=self.api_url,
+                entity_type=entity_type,
+                entity_id=entity_id,
+                asset_id=None,
+            )
+            + "/directory/upload"
+        )
+        context = self._required_user_context(override_context=project_context)
+
+        paths = {Path(k): Path(v) for k, v in paths.items()}
+
+        return core.upload_asset_directory(
+            url=url,
+            name=name,
+            paths=paths,
+            metadata=metadata,
+            label=label,
+            project_context=context,
+            http_client=self._http_client,
+            token=self._token_manager.get_token(),
+        )
+
+    def list_directory(
+        self,
+        *,
+        entity_id: ID,
+        entity_type: type[Identifiable],
+        asset_id: ID,
+        project_context: ProjectContext | None = None,
+    ) -> DetailedFileList:
+        """List directory existing entity's endpoint from a directory path."""
+        url = (
+            route.get_assets_endpoint(
+                api_url=self.api_url,
+                entity_type=entity_type,
+                entity_id=entity_id,
+                asset_id=asset_id,
+            )
+            + "/list"
+        )
+        context = self._required_user_context(override_context=project_context)
+        return core.list_directory(
+            url=url,
+            project_context=context,
+            http_client=self._http_client,
+            token=self._token_manager.get_token(),
+        )
+
+    def download_directory(
+        self,
+        *,
+        entity_id: ID,
+        entity_type: type[Identifiable],
+        asset_id: ID,
+        output_path: os.PathLike,
+        project_context: ProjectContext | None = None,
+        ignore_directory_name: bool = False,
+    ) -> list[Path]:
+        """List directory existing entity's endpoint from a directory path."""
+        output_path = Path(output_path)
+
+        if output_path.exists() and output_path.is_file():
+            raise EntitySDKError(f"{output_path} exists and is a file")
+        output_path.mkdir(parents=True, exist_ok=True)
+
+        context = self._optional_user_context(override_context=project_context)
+
+        asset = None
+        if not ignore_directory_name:
+            asset_endpoint = route.get_assets_endpoint(
+                api_url=self.api_url,
+                entity_type=entity_type,
+                entity_id=entity_id,
+                asset_id=asset_id,
+            )
+            asset = core.get_entity(
+                asset_endpoint,
+                entity_type=Asset,
+                project_context=context,
+                http_client=self._http_client,
+                token=self._token_manager.get_token(),
+            )
+
+            output_path /= asset.path
+
+        contents = self.list_directory(
+            entity_id=entity_id,
+            entity_type=entity_type,
+            asset_id=asset_id,
+            project_context=project_context,
+        )
+
+        paths = []
+        for path in contents.files:
+            paths.append(
+                self.download_file(
+                    entity_id=entity_id,
+                    entity_type=entity_type,
+                    asset_id=asset if asset else asset_id,
+                    output_path=output_path / path,
+                    asset_path=path,
+                    project_context=context,
+                )
+            )
+
+        return paths
+
     def download_content(
         self,
         *,
         entity_id: ID,
         entity_type: type[Identifiable],
-        asset_id: str,
+        asset_id: ID,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> bytes:
         """Download asset content.
 
@@ -325,7 +408,6 @@ class Client:
             entity_type: Type of the entity.
             asset_id: Id of the asset.
             project_context: Optional project context.
-            token: Authorization access token.
 
         Returns:
             Asset content in bytes.
@@ -339,13 +421,12 @@ class Client:
             )
             + "/download"
         )
-        token = self._get_token(override_token=token)
         context = self._optional_user_context(override_context=project_context)
         return core.download_asset_content(
             url=url,
-            token=token,
             project_context=context,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def download_file(
@@ -353,40 +434,116 @@ class Client:
         *,
         entity_id: ID,
         entity_type: type[Identifiable],
-        asset_id: str,
+        asset_id: ID | Asset,
         output_path: os.PathLike,
+        asset_path: os.PathLike | None = None,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
-    ) -> None:
+    ) -> Path:
         """Download asset file to a file path.
 
         Args:
             entity_id: Id of the entity.
             entity_type: Type of the entity.
             asset_id: Id of the asset.
-            output_path: Path to save the file to.
+            output_path: Either be a file path to write the file to or an output directory.
+            asset_path: for asset directories, the path within the directory to the file
             project_context: Optional project context.
-            token: Authorization access token.
+
+        Returns:
+            Output file path.
         """
-        url = (
-            route.get_assets_endpoint(
-                api_url=self.api_url,
-                entity_type=entity_type,
-                entity_id=entity_id,
-                asset_id=asset_id,
-            )
-            + "/download"
-        )
-        token = self._get_token(override_token=token)
         context = self._optional_user_context(override_context=project_context)
-        return core.download_asset_file(
-            url=url,
-            token=token,
-            project_context=context,
-            output_path=Path(output_path),
-            http_client=self._http_client,
+        asset_endpoint = route.get_assets_endpoint(
+            api_url=self.api_url,
+            entity_type=entity_type,
+            entity_id=entity_id,
+            asset_id=asset_id if isinstance(asset_id, ID) else asset_id.id,
         )
 
+        if isinstance(asset_id, ID):
+            asset = core.get_entity(
+                asset_endpoint,
+                entity_type=Asset,
+                project_context=context,
+                http_client=self._http_client,
+                token=self._token_manager.get_token(),
+            )
+        else:
+            asset = asset_id
+
+        path: Path = Path(output_path)
+        if asset.is_directory:
+            if not asset_path:
+                raise EntitySDKError("Directory from directories require an `asset_path`")
+        else:
+            if asset_path:
+                raise EntitySDKError("Cannot pass `asset_path` to non-directories")
+
+            path = (
+                path / asset.path
+                if path.is_dir()
+                else validate_filename_extension_consistency(path, Path(asset.path).suffix)
+            )
+        create_intermediate_directories(path)
+        return core.download_asset_file(
+            url=f"{asset_endpoint}/download",
+            project_context=context,
+            asset_path=asset_path,
+            output_path=path,
+            http_client=self._http_client,
+            token=self._token_manager.get_token(),
+        )
+
+    def download_assets(
+        self,
+        entity_or_id: Entity | tuple[ID, type[Entity]],
+        *,
+        selection: dict[str, Any] | None = None,
+        output_path: Path,
+        project_context: ProjectContext | None = None,
+    ) -> IteratorResult:
+        """Download assets."""
+
+        def _download_entity_asset(asset):
+            if asset.is_directory:
+                raise NotImplementedError("Downloading asset directories is not supported yet.")
+            else:
+                path = self.download_file(
+                    entity_id=entity.id,
+                    entity_type=type(entity),
+                    asset_id=asset.id,
+                    output_path=output_path,
+                    project_context=context,
+                )
+
+            return DownloadedAsset(
+                asset=asset,
+                output_path=path,
+            )
+
+        context = self._optional_user_context(override_context=project_context)
+        if isinstance(entity_or_id, tuple):
+            entity_id, entity_type = entity_or_id
+            entity = self.get_entity(
+                entity_id=entity_id,
+                entity_type=entity_type,
+                project_context=context,
+            )
+        else:
+            entity = entity_or_id
+
+        if not issubclass(type(entity), Entity):
+            raise EntitySDKError(f"Type {type(entity)} has no assets.")
+
+        # make mypy happy as it doesn't get the correct type :(
+        entity = cast(Entity, entity)
+
+        if not entity.assets:
+            raise EntitySDKError(f"Entity {entity.id} ({entity.name}) has no assets.")
+
+        assets = filter_assets(entity.assets, selection) if selection else entity.assets
+        return IteratorResult(map(_download_entity_asset, assets))
+
     def delete_asset(
         self,
         *,
@@ -394,7 +551,6 @@ class Client:
         entity_type: type[Identifiable],
         asset_id: ID,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Asset:
         """Delete an entity's asset."""
         url = route.get_assets_endpoint(
@@ -403,13 +559,12 @@ class Client:
             entity_id=entity_id,
             asset_id=asset_id,
         )
-        token = self._get_token(override_token=token)
         context = self._required_user_context(override_context=project_context)
         return core.delete_asset(
             url=url,
-            token=token,
             project_context=context,
             http_client=self._http_client,
+            token=self._token_manager.get_token(),
         )
 
     def update_asset_file(
@@ -423,7 +578,6 @@ class Client:
         file_name: str | None = None,
         file_metadata: dict | None = None,
         project_context: ProjectContext | None = None,
-        token: str | None = None,
     ) -> Asset:
         """Update an entity's asset file.
 
@@ -434,7 +588,6 @@ class Client:
             entity_type=entity_type,
             asset_id=asset_id,
             project_context=project_context,
-            token=token,
         )
         return self.upload_file(
             entity_id=entity_id,
@@ -444,5 +597,4 @@ class Client:
             file_name=file_name,
             file_metadata=file_metadata,
             project_context=project_context,
-            token=token,
         )
