@@ -7,13 +7,18 @@ from entitysdk.models import (
     IonChannelModel,
     License,
     MEModel,
+    MEModelCalibrationResult,
     MTypeClass,
     Organization,
     Person,
     ReconstructionMorphology,
     Role,
+    SingleNeuronSimulation,
+    SingleNeuronSynaptome,
+    SingleNeuronSynaptomeSimulation,
     Species,
     Strain,
+    ValidationResult,
 )
 
 
@@ -33,6 +38,11 @@ from entitysdk.models import (
         EModel,
         MEModel,
         ElectricalCellRecording,
+        ValidationResult,
+        MEModelCalibrationResult,
+        SingleNeuronSimulation,
+        SingleNeuronSynaptomeSimulation,
+        SingleNeuronSynaptome,
     ],
 )
 def test_is_searchable(entity_type, client):
